// -* coding: utf-8; -*

var MIM = {};

// URL of the input method server.
MIM.server = "http://www.m17n.org/common/mim-js",
// Boolean flag to tell if MIM is active or not.
MIM.enabled = true;
// Boolean flag to tell if MIM is running in debug mode or not.
MIM.debug = false;
// List of registered input methods.
MIM.list = new Array ();
// Currently selected input method.
MIM.current_im = false;

MIM.im = function (lang, name, filename)
{
  this.status = 0; /* 0: not-yet-loaded, 1:loading, 2:loaded, -1:error */
  this.url = MIM.server + "/" + filename;
  this.lang = lang;
  this.name = name;
  this.keymap = false;
  this.body = null;

  function add_keystring (map, keystring, str)
  {
    var i, c;
    var newmap;
    var intermediate_string = "";

    for (i = 0; i < keystring.length; i++)
    {
      c = keystring.charAt (i);
      if (c in map)
        {
	  map = map[c];
	  if ('_target_text' in map)
	    intermediate_string = map['_target_text'];
	  else
	    intermediate_string += c;
	}
      else
        {
	  newmap = new Array ();
	  map[c] = newmap;
	  map['_has_child'] = true;
	  map = newmap;
	  intermediate_string += c;
	  map['_target_text'] = intermediate_string;
	}
    }
    map['_target_text'] = str;
  }

  this.lookup = function (keyseq, limit)
  {
    var map = this.keymap;

    if (limit > keyseq.length)
      limit = keyseq.length;
    for (var i = 0; i < limit; i++)
      {
	var c = keyseq[i];
	if (! (c in map))
	  return i;
	map = map[c];
      }
    return map;
  }

  this.load_map = function (mapdef)
  {
    this.keymap = new Array ();
    for (var keystring in mapdef)
      add_keystring (this.keymap, keystring, mapdef[keystring]);
  }

  this.load_map_node = function ()
  {
    this.keymap = new Array ();
    var maps = this.body.getElementsByTagName ('map');
    var map = maps[0];
    var rules = map.getElementsByTagName ('rule');
    for (var i = 0; i < rules.length; i++)
      {
	var rule = rules[i];
	var keyseq_elm = MIM.first_element (rule);
	var keystring = keyseq_elm.attributes[0].nodeValue;
	var insert_elm = MIM.next_element (rule);
	var str = insert_elm.attributes[0].nodeValue;
	add_keystring (this.keymap, keystring, str);
      }
  }
};

MIM.error_return = function (msg, ret)
{
  alert (msg);
  return ret;
}

MIM.first_element = function (node)
{
  node.mim_index = 0;
  return MIM.next_element (node);
}

MIM.next_element = function (node)
{
  var element = node.childNodes[node.mim_index++];
  while (element && element.nodeType != 1)
    element = node.childNodes[node.mim_index++];
  return element;
}

MIM.check_map = function (im, map)
{
  var rules = map.getElementsByTagName ('rule');
  var len = rules.length;
  
  for (var i = 0; i < len; i++)
    {
      var rule = rules[i];
      var elm = MIM.first_element (rule);

      if (!elm || elm.nodeName != 'keyseq')
	return false;
      while ((elm = MIM.next_element (rule)))
	if (elm.nodeName != 'insert')
	  return false;
    }
  return true;
}

MIM.check_state = function (im, state)
{
  var branches = state.getElementsByTagName ('branch');
  var len = branches.length;

  for (var i = 0; i < len; i++)
    {
      var branch = branches[i];
      var elm = MIM.first_element (branch);

      if (elm)
	return false;
    }
  return true;
}

MIM.parse = function (im)
{
  var maps = im.body.getElementsByTagName ('map');
  var states = im.body.getElementsByTagName ('state');
  var str = "";
  var i;

  if (! maps || maps.length == 0)
    MIM.error_return ('No map', false);
  if (! states)
    MIM.error_return ('No state', false);
  for (i = 0; i < maps.length; i++)
    if (! MIM.check_map (im, maps[i]))
      MIM.error_return ('Unsupported directive in map', false);
  for (var i = 0; i < states.length; i++)
    if (! MIM.check_state (im, states[i]))
      MIM.error_return ('Unsupported directive in state', false);
  im.load_map_node ();
  return true;
}

MIM.register = function (lang, name, url)
{
  var im = new MIM.im (lang, name, url);
  if (! (lang in MIM.list))
    MIM.list[lang] = new Array ();
  MIM.list[lang][name] = im;
  return im;
};

MIM.find = function (lang, name)
{
  if (! (lang in MIM.list))
    return false;
  if (! (name in MIM.list[lang]))
    return false;
  return MIM.list[lang][name];
};

MIM.load_async = function (im)
{
  var obj = (window.XMLHttpRequest ? new XMLHttpRequest ()
	     : window.ActiveXObject ? new ActiveXObject ("Msxml2.XMLHTTP")
	     : null);

  if (! obj)
    alert ("XMLHttpRequest not supported");
  obj.open ('GET', im.url, true);
  im.status = 1; /* loading */
  obj.onreadystatechange = function () { 
    if (obj.readyState == 4)
      {
	try {
	  eval (obj.responseText);
	  im.status = 2; /* loaded */
	} catch (e) {
	  alert ("load error:" + e.message + " at " + e.lineNumber
		 + " " + obj.responseText);
	  im.status = -1; /* load fail */
	}
      }
  };
  obj.send (null);
  return im;
};

MIM.load_sync = function (im)
{
  var obj = (window.XMLHttpRequest ? new XMLHttpRequest ()
	     : window.ActiveXObject ? new ActiveXObject ("Msxml2.XMLHTTP")
	     : null);

  if (! obj)
    alert ("XMLHttpRequest not supported");
  if (true)
    {
      obj.open ('GET', im.url, false);
      obj.send ("");  
      eval (obj.responseText);
      if (! im.keymap)
	alert ("obj.responseText");
      return im;
    }
  else
    {
      obj.open ('GET', 'latn-post.mimx', false);
      obj.overrideMimeType ('text/xml');
      obj.send ("");
      im.body = obj.responseXML;
      document.AnXml = im.body;
      if (MIM.parse (im))
	return im;
      alert (im.parse_error);
    }
  return false;
};

MIM.load = function (im)
{
  var s = document.createElement ('script');

  s.charset = 'UTF-8';
  s.src = im.url;
  document.body.appendChild (s);
  document.body.removeChild (s);
  if (! im.keymap)
    alert ("load fail");
  im.status = 2;
  return im;
};

MIM.add_event_listener
  = (window.addEventListener
     ? function (target, type, listener) {
       target.addEventListener (type, listener, false);
     }
     : window.attachEvent
     ? function (target, type, listener) {
       target.attachEvent ('on' + type,
			   function() {
			     listener.call (target, window.event);
			   });
     }
     : function (target, type, listener) {
       target['on' + type]
	 = function (e) { listener.call (target, e || window.event); };
     });

(function () {
  var keys = new Array ();
  keys[0x09] = 'tab';
  keys[0x08] = 'backspace';
  keys[0x0D] = 'return';
  keys[0x1B] = 'escape';
  keys[0x20] = 'space';
  keys[0x21] = 'pageup';
  keys[0x22] = 'pagedown';
  keys[0x23] = 'end';
  keys[0x24] = 'home';
  keys[0x25] = 'left';
  keys[0x26] = 'up';
  keys[0x27] = 'right';
  keys[0x28] = 'down';
  keys[0x2D] = 'insert';
  keys[0x2E] = 'delete';
  for (var i = 1; i <= 12; i++)
    keys[111 + i] = "f" + i;
  keys[0x90] = "numlock";
  keys[0xF0] = "capslock";
  MIM.special_key = keys;
}) ();

MIM.decode_key = function (event)
{
  var key = ((event.type == 'keydown' || event.keyCode) ? event.keyCode
	     : event.charCode ? event.charCode
	     : false);
  if (! key)
    return false;
  if (event.type == 'keydown')
    {
      key = MIM.special_key[key];
      if (! key)
	return false;
      if (event.shiftKey) key = "S-" + key ;
      }
  else
    key = String.fromCharCode (key);
  if (event.altKey) key = "A-" + key ;
  if (event.ctrlKey) key = "C-" + key ;
  return key;
};

MIM.debug_print = function (event, ic)
{
  if (! MIM.debug)
    return;
  if (! MIM.debug_nodes)
    {
      MIM.debug_nodes = new Array ();
      MIM.debug_nodes['keydown'] = document.getElementById ('keydown');
      MIM.debug_nodes['keypress'] = document.getElementById ('keypress');
      MIM.debug_nodes['status0'] = document.getElementById ('status0');
      MIM.debug_nodes['status1'] = document.getElementById ('status1');
      MIM.debug_nodes['keyseq0'] = document.getElementById ('keyseq0');
      MIM.debug_nodes['keyseq1'] = document.getElementById ('keyseq1');
      MIM.debug_nodes['range0'] = document.getElementById ('range0');
      MIM.debug_nodes['range1'] = document.getElementById ('range1');
    }
  var target = event.target;
  var code = event.keyCode;
  var ch = event.type == 'keydown' ? 0 : event.charCode;
  var key = MIM.decode_key (event);
  var keyseq = "";
  var index;

  MIM.debug_nodes[event.type].innerHTML = "" + code + "/" + ch + " : " + key;
  index = (event.type == 'keydown' ? '0' : '1');
  MIM.debug_nodes['status' + index].innerHTML = ic.im.status;
  for (var i = 0; i < ic.keyseq.length; i++)
    keyseq += ic.keyseq[i];
  MIM.debug_nodes['keyseq' + index].innerHTML
    = keyseq + ":" + ic.keyseq.length;
  MIM.debug_nodes['range' + index].innerHTML
    = "" + ic.range[0] + ":" + ic.range[1];
};

MIM.get_range = function (target, range)
{
  if (target.selectionStart != null) // for Mozilla
    {
      range[0] = target.selectionStart;
      range[1] = target.selectionEnd;
    }
  else				// for IE
    {
      var r = document.selection.createRange ();
      var rr = r.duplicate ();

      rr.moveToElementText (target);
      rr.setEndPoint ('EndToEnd', range);
      range[0] = rr.text.length - r.text.length;
      range[1] = rr.text.length;
    }
}

MIM.set_caret = function (target, pos)
{
  if (target.selectionStart != null) // Mozilla
    {
      target.focus ();
      target.setSelectionRange (pos, pos);
    }
  else				// IE
    {
      var range = target.createTextRange ();
      range.move ('character', pos);
      range.select ();
    }
};

MIM.ic = function (im, target)
{
  this.im = im;
  this.target = target;
  this.key = false;
  this.keyseq = new Array ();
  this.range = new Array (0, 0);
  return this;
};

MIM.ic.prototype.reset = function ()
{
  this.key = false;
  while (this.keyseq.length > 0)
    this.keyseq.pop ();
};

MIM.ic.prototype.check_range = function ()
{
  var from = this.range[0], to = this.range[1];

  MIM.get_range (this.target, this.range);
  if (this.range[0] != this.range[1] || to != this.range[1])
    this.reset ();
  else
    this.range[0] = from;
};

MIM.insert = function (ic, insert)
{
  var text = ic.target.value;
  ic.target.value = (text.substring (0, ic.range[0])
		     + insert
		     + text.substring (ic.range[1]));
  ic.range[1] = ic.range[0] + insert.length;
  MIM.set_caret (ic.target, ic.range[1]);
};

function keyseq_string (keyseq)
{
  var str = "";
  for (var i = 0; i < keyseq.length; i++)
    str += keyseq[i];
  return str;
}

MIM.handle_keyseq = function (event, ic)
{
  var map = ic.im.lookup (ic.keyseq, 1000);
  if (map instanceof Array)
    {
      MIM.insert (ic, map['_target_text']);
      if (! ('_has_child' in map))
	{
	  ic.reset ();
	  ic.range[0] = ic.range[1];
	}
      event.preventDefault ();
      //document.getElementById ('text').value
      //= keyseq_string (ic.keyseq) + " handled";
    }
  else if (map > 0)
    {
      MIM.insert (ic, ic.im.lookup (ic.keyseq, map)['_target_text']);
      while (map > 0)
	{
	  ic.keyseq.shift ();
	  map--;
	}
      ic.range[0] = ic.range[1];
      if (ic.keyseq.length > 0)
	MIM.handle_keyseq (event, ic);
    }
  else
    {
      ic.reset ();
      //document.getElementById ('text').value
      //= keyseq_string (ic.keyseq) + " unhandled";
    }
};

MIM.reset_ic = function (event)
{
  var ic = event.target.mim_ic;
  if (ic)
    ic.reset ();
};

MIM.keydown = function (event)
{
  var target = event.target;
  if (! (target.type == "text" || target.type == "textarea"))
    return;

  var ic = target.mim_ic;
  if (! ic || ic.im != MIM.current_im)
    {
      ic = new MIM.ic (MIM.current_im, target);
      target.mim_ic = ic;
      MIM.add_event_listener (target, 'blur', MIM.reset_ic);
    }
  if (ic.im.status < 0)
    return;
  ic.check_range ();
  MIM.debug_print (event, ic);
  ic.key = MIM.decode_key (event);
};

MIM.keypress = function (event)
{
  if (! (event.target.type == "text" || event.target.type == "textarea"))
    return;

  var ic = event.target.mim_ic;
  var i;

  try {
    if (ic.im.status < 0)
      return;
    if (! ic.key)
      ic.key = MIM.decode_key (event);
    if (! ic.key)
      {
	ic.reset ();
	return;
      }
    ic.keyseq.push (ic.key);
    if (ic.im.status == 1) // Still loading.
      return;
    MIM.handle_keyseq (event, ic);
  } finally {
    MIM.debug_print (event, ic);
  }
  return;
};

MIM.select_im = function (event)
{
  var target = event.target.parentNode;
  while (target.tagName != "SELECT")
    target = target.parentNode;
  var idx = 0;
  var im = false;
  for (var lang in MIM.list)
    for (var name in MIM.list[lang])
      if (idx++ == target.selectedIndex)
	{
	  im = MIM.list[lang][name];
	  break;
	}
  document.getElementsByTagName ('body')[0].removeChild (target);
  target.target.focus ();
  if (im && im != MIM.current_im)
    MIM.current_im = MIM.load_sync (im);
};

MIM.destroy_menu = function (event)
{
  if (event.target.tagName == "SELECT")
    document.getElementsByTagName ('body')[0].removeChild (event.target);
};

MIM.select_menu = function (event)
{
  var target = event.target;

  if (! ((target.type == "text" || target.type == "textarea")
	 && event.which == 1 && event.ctrlKey))
    return;

  var sel = document.createElement ('select');
  sel.onclick = MIM.select_im;
  sel.onmouseout = MIM.destroy_menu;
  sel.style.position='absolute';
  sel.style.left = (event.clientX - 10) + "px";
  sel.style.top = (event.clientY - 10) + "px";
  sel.target = target;
  var idx = 0;
  for (var lang in MIM.list)
    for (var name in MIM.list[lang])
      {
	var option = document.createElement ('option');
	var imname = lang + "-" + name;
	option.appendChild (document.createTextNode (imname));
	option.value = imname;
	sel.appendChild (option);
	if (MIM.list[lang][name] == MIM.current_im)
	  sel.selectedIndex = idx;
	idx++;
      }
  sel.size = idx;
  document.getElementsByTagName ('body')[0].appendChild (sel);
};

MIM.textinput = function (event)
{
  var str = ''
  var changed = false;
  for (var i = 0; i < event.data.length; i++)
    {
      if (event.data.charAt (i) == 'あ')
	{
	  str += 'ア'; changed = true;
	}
      else
	str += event.data.charAt (i);
    }
  alert (str);
  if (changed)
    {
      var e = document.createEvent ('TextEvent');
      e.initTextEvent ('textInput', event.canBuggle, event.cancelable, event.view,
		       str, event.inputMode);
      document.getElementById ('text').value = str;
      event.target.dispatchEvent (e);
    }
}

MIM.init = function ()
{
  MIM.add_event_listener (window, 'keydown', MIM.keydown);
  MIM.add_event_listener (window, 'keypress', MIM.keypress);
  MIM.add_event_listener (window, 'mousedown', MIM.select_menu);
  MIM.add_event_listener (window, 'textInput', MIM.textinput);
  if (window.location == 'http://localhost/mim/index.html')
    MIM.server = 'http://localhost/mim';
  MIM.current_im = MIM.register ('latin', 'post', 'latn-post.js');
  MIM.register ('th', 'kesmanee', 'th-kesmanee.js');
  MIM.load_sync (MIM.current_im);
};

MIM.init_debug = function ()
{
  MIM.debug = true;
  MIM.init ();
};
